// Student: Maxim Alexandru
// Grupa: 332CA

#define	 F_CPU	16000000UL
#include <avr/io.h>
#include <avr\interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdint.h>
#include <avr/eeprom.h>
#define LCD_INSTR_DDRAM         0x80    // prima adresa din DDRAM
#define LCD_INSTR_CGRAM         0x40    // prima adresa din CGRAM

// Biblioteca pentru LCD
// https://community.atmel.com/projects/ili9341-library-drive-22-tft-displayderived-adafruit-tft-library-ili9340-type-controller

#include "ili9341.h"
#include "ili9341gfx.h"

#define LENWORD 14
#define SPEAKER_PORT PORTD
#define SPEAKER_DDR DDRD
#define SPEAKER_PIN 5

extern uint16_t vsetx,vsety,vactualx,vactualy,isetx,isety,iactualx,iactualy;
static FILE mydata = FDEV_SETUP_STREAM(ili9341_putchar_printf, NULL, _FDEV_SETUP_WRITE);

int cursorx;
int cursory;
int lvl_correct;
int rectangle_x = 35;
int rectangle_y = 35;
int circle_diffx = 17;
int circle_diffy = 17;
int diff_rectanglex = 40;
int shapes_revealed = 0;
int shapes_revealed_for_a_game;
int shape2_x = -1, shape1_x = -1, shape1_y = -1, shape2_y = -1;
int cursor_menu_x = 30 + 50, cursor_menu_y = 50 + 20;
int select_x = 30 + 50, select_y = 45 + 20;
int select_xc = 80, select_yc = 65;
volatile int cursor;
volatile int l = 0;
volatile int in_game = 0;
int number_squares = 0;
int lvl_x, lvl_y;
volatile int castigat;
volatile int nr_moves;
volatile int lvls;
volatile int start_again;
volatile int first_time = 1;

char *shapes[2];
int id = 0;
int ids[5] = {0};
int scores[5] = {0.0};
int nr_lvls[5] = {0};
volatile int seconds = 0;

int **revealed;

// r1, c1, r2, c2 sunt coordonatele simbolurilor care au fost dezvaluite de catre jucator
// r, c coordonatele din joc la care se afla jucatorul
volatile int r1, r2, c1, c2, r, c;
int startx = 10, startxc = 10;
int starty = 50, startyc = 50;
int correct = 0;


// functie ce trimite semnale de o anumita frecventa si anumita durata
void PLAYNOTE(float duration, float frequency);

// pool-ul de posibilitati din care se alege aleator
char vector_chars1[4][7] = {{'a', 0x2d, 0xd2, 'd', '.', 0x8b, 'g'}, {0x8b, 'c', '~', 0xaa, 0x9e, 'g', '+'}, {0x3f, '*', 'e', '^', 0x4d, 0x8f, 0x8b}, {0x3c, 0x5e, 'a', 'b', 'c', 'f', 'g'}};
char vector_chars2[4][7] = {{0x2d, 0xd2, 'd', '.', 'a', 0x8b, 'g'}, {0x8b, 0x9e, 0xaa, '~', 'c', '+', 'g'}, {'^', 0x4d, 0x3f, '*', 'e', 0x8b, 0x8f}, {'c', 'b', 'a', 0x5e, 0x3c, 'f', 'g'}};
char vector_not[4][7] = {{'x', 0x2d, 0xd2, 'd', '.', 0x8b, 'g'}, {0x8b, 'c', '-', 0xaa, 0x9e, 'g', '+'}, {0x3f, '+', 'e', '^', 0x4d, 'g', 'f'}, {0x3c, 0xcc, 'a', 'b', 'c', 'f', 'g'}};
char *print_row1, *print_row2;
volatile int times[5] = {0};


// muzica star wars
// luata din: https://gist.github.com/nicksort/4736535 
const int cx = 261;
const int d = 294;
const int e = 329;
const int f = 349;
const int g = 391;
const int gS = 415;
const int a = 440;
const int aS = 455;
const int b = 466;
const int cH = 523;
const int cSH = 554;
const int dH = 587;
const int dSH = 622;
const int eH = 659;
const int fH = 698;
const int fSH = 740;
const int gH = 784;
const int gSH = 830;
const int aH = 880;
 
const int buzzerPin = 8;
const int ledPin1 = 12;
const int ledPin2 = 13;
 
int counter = 0;
 
 void firstSection()
{
  PLAYNOTE(500, a);
  PLAYNOTE(500, a);    
  PLAYNOTE(500, a);
  PLAYNOTE(350, f);
  PLAYNOTE(150, cH);  
  PLAYNOTE(500, a);
  PLAYNOTE(350, f);
  PLAYNOTE(150, cH);
  PLAYNOTE(650, a);
 
  _delay_ms(500);
 
  PLAYNOTE(500, eH);
  PLAYNOTE(500, eH);
  PLAYNOTE(500, eH);  
  PLAYNOTE(350, fH);
  PLAYNOTE(150, cH);
  PLAYNOTE(500, gS);
  PLAYNOTE(350, f);
  PLAYNOTE(150, cH);
  PLAYNOTE(650, a);
 
  _delay_ms(500);
}
 
 
void loop()
{
  firstSection();
}

// functie care este apelata cand jucatorul a ales un simbol
// verifica daca 2 simboluri au fost alese si le compara
// daca cele 2 simboluri sunt identice, acestea raman dezvaluite, se aude un sunet de la difuzor
// daca nu sunt, sunt ascunse de patraturi negre
void reveal_shape(int sx, int sy) {
	int re = 0;
	if(shapes_revealed < 2 && !revealed[r][c]) {
		if(shape1_x != -1) {
			shape2_x = sx;
			shape2_y = sy;
			r2 = r;
			c2 = c;
		}
		else{
			shape1_x = sx;
			shape1_y = sy;
			r1 = r;
			c1 = c;
			re = 1;
		}
			revealed[r][c] = 1;
			shapes_revealed++;
			ili9341_fillrect(sx, sy, rectangle_x, rectangle_y, RED);
			_delay_ms(1);

			ili9341_setcursor(sx+10, sy+10);
			printf("%c", shapes[r][c]);
	}
	if(shapes_revealed >= 2) {
		if(shapes[r1][c1] != shapes[r2][c2]) {
			_delay_ms(1000);
			ili9341_fillrect(shape1_x, shape1_y, rectangle_x, rectangle_y, BLACK);
			ili9341_fillrect(shape2_x, shape2_y, rectangle_x, rectangle_y, BLACK);
			_delay_ms(10);
			ili9341_Draw_Hollow_Circle(shape2_x + circle_diffx, shape2_y + circle_diffy, 10, GREEN);
			revealed[r1][c1] = 0;
			revealed[r2][c2] = 0;
		}
		else {
			PLAYNOTE(200,700);
			correct++;
		}
		shape2_x = -1;
		shape2_y = -1;
		shape1_x = -1;
		shape1_y = -1;
		c2 = 0;
		c1 = 0;
		r1 = 0;
		r2 = 0;
		shapes_revealed = 0;
	
	}

}

// timer la fiecare secunda
// afiseaza un cronometru pentru jucator
// afiseaza si ID-ul jucatorului
ISR(TIMER1_COMPA_vect)
{	if(first_time){
		first_time = 0;
		_delay_ms(300);
	}
	int hr, min, sec;
	ili9341_settextsize(2.5);
	ili9341_setcursor(10, 200);
	if(seconds>3600){
		min = seconds/60;
		sec = seconds%60;
		hr = min/60;
		min = min%60;
		printf("%d:%d:%d",hr,min,sec);
	}
	else{
		min = seconds/60;
		sec = seconds%60;
		printf("%d:%d",min,sec);
	}
	seconds++;
	ili9341_setcursor(10,5);
	printf("Player's id: %d", id);
}



// functie pentru a misca cursorul in dreapta in timpul unui nivel
void try_right(int rectangles) {
	if(startx <= rectangles*rectangle_x) {
		
		if(!revealed[r][c]) {
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, BLACK);
			_delay_ms(1);
		}
		else {
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, RED);
			_delay_ms(1);
			ili9341_setcursor(startx+10, starty+10);
			printf("%c", shapes[r][c]);
		}
		startx += diff_rectanglex;
			
		ili9341_Draw_Hollow_Circle(startx + circle_diffx, starty + circle_diffy, 10, GREEN);
		_delay_ms(1);
		c++;
	}
}

// functie pentru a misca cursorul in stanga in timpul unui nivel
void try_left(int start) {
	if(startx > start) {
		
		if(!revealed[r][c]){
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, BLACK);
			_delay_ms(1);
		}
		else {
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, RED);
			_delay_ms(1);
			ili9341_setcursor(startx+10, starty+10);
			printf("%c", shapes[r][c]);
		}
		startx -= diff_rectanglex;
			
		ili9341_Draw_Hollow_Circle(startx + circle_diffx, starty + circle_diffy, 10, GREEN);
		_delay_ms(1);
		c--;
	}
}

// functie pentru a misca cursorul in sus in timpul unui nivel
void try_up(int start) {
	if(starty > start) {
	
		if(!revealed[r][c]) {
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, BLACK);
			_delay_ms(1);
		}
		else {
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, RED);
			_delay_ms(1);
			ili9341_setcursor(startx+10, starty+10);
			printf("%c", shapes[r][c]);
		}
		starty -= diff_rectanglex;
			
		ili9341_Draw_Hollow_Circle(startx + circle_diffx, starty + circle_diffy, 10, GREEN);
		_delay_ms(1);
		r--;
	}
}

// functie pentru a misca cursorul in jos in timpul unui nivel
void try_down(int start) {
	if(starty <= start) {
		if(!revealed[r][c]){
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, BLACK);
			_delay_ms(1);
		}
		else {
			ili9341_fillrect(startx, starty, rectangle_x, rectangle_y, RED);
			_delay_ms(1);
			ili9341_setcursor(startx+10, starty+10);
			printf("%c", shapes[r][c]);
		}
		starty += diff_rectanglex;
			
		ili9341_Draw_Hollow_Circle(startx + circle_diffx, starty + circle_diffy, 10, GREEN);
		_delay_ms(1);
		r++;
	}
}


// intrerupere cand se apasa pe butoane
// pentru fiecare buton apasat exista o functie care misca cursorul in timpul unui nivel
// sau pentru butonul care alege un patrat negru, e apelata functia reveal_shape
// daca jucatorul a castigat, se incrementeaza un flag si se trece la meniu
ISR(PCINT0_vect)
{
	if(in_game && (PINA & (1 << PA2)) == 0) {
		reveal_shape(startx, starty);
		ili9341_settextsize(2.5);
		ili9341_setcursor(10, 170);
		nr_moves++;
		printf("Pairs remained: %d", lvl_correct - correct);
		if(lvl_correct == correct) {
			castigat = 1;
			in_game = 0;
			lvls++;
			cursor = 0;
		}
	}
	
			if (in_game && (PINA & (1 << PA0)) == 0) {
				try_right(number_squares);
			}
			// Move Left PA4
			if (in_game && (PINA & (1 << PA4)) == 0) {
				try_left(lvl_x);
			}
			
			if (in_game && (PINA & (1 << PA1)) == 0) {
				try_up(lvl_y);
			}
			
			if (in_game && (PINA & (1 << PA3)) == 0) {
				try_down(lvl_y);
			}

}

// functia deseneaza nivelul pentru ce parametrii primeste si cursorul la inceput
void draw_board(int startx_v, int starty_v, int rectangles, int start) {
	int i =0;
	for(i = 0; i < rectangles; i++) {
		ili9341_fillrect(startx_v, starty_v, rectangle_x, rectangle_y, BLACK);
		_delay_ms(1);
		
		if(i == 0 && start == 1){
			ili9341_Draw_Hollow_Circle(startx_v + circle_diffx, starty_v + circle_diffy, 10, GREEN);
			_delay_ms(1);
		}
		startx_v += diff_rectanglex;
	
	}

}

// functia leaga pool-ul de vectori din care se alege random o configuratie
// de matricea shapes care o sa fie folosita pentru afisare si verificare
// am generat numeere random folosind functii eeprom_read_dword
// am gasit pe internet un post care explica cum trebuie folosite
void bind_prints(int l) {
	_delay_ms(2);
	ili9341_settextcolour(YELLOW,RED);
	_delay_ms(2);
	ili9341_settextsize(3);
	_delay_ms(2);
	
	
	uint32_t state;
    static uint32_t sstate;
    state = eeprom_read_dword(&sstate);
	if (state == 0xffffffUL)
        state = 0xDEADBEEFUL;
	srand(state);
	eeprom_write_dword(&sstate, random());
	
	int rand_number = (int)(rand() % 4);
	ili9341_setcursor(10, 150);
	if(l < 2) {
		shapes[0] = vector_chars1[rand_number];
		shapes[1] = vector_chars2[rand_number];
	}
	
	else {
		shapes[0] = vector_chars1[rand_number];
		shapes[1] = vector_not[rand_number];
	}
}

// tabela de scor hall_of_fame, afiseaza un scor pentru un jucator
// daca aceasta este selectata, toate datele despre vechiul jucator sunt sterse
// daca se apasa pe butonul din mijloc te trimite in meniu pentru a incepe alt jucator
void hall_of_fame() {
	int score = nr_moves * lvls * 100;
	ili9341_clear(RED);//fill screen with a colour
	nr_lvls[id] = lvls;
	nr_moves = 0;
	ids[id] = id;
	times[id] += seconds;
	ili9341_setcursor(70, 10);
	printf("HALL OF FAME");
	lvls = 0;
	scores[id] = score;
	if(id < 4) {
		id++;
	}
	int i;
	ili9341_setcursor(10, 40);
	printf("ID   Score  Levels  Time");
	for(i = 0; i < 5; i++){
		ili9341_setcursor(10, 60 + i*30);
		printf("%d     %d     %d    %d", ids[i], scores[i], nr_lvls[i], times[i]);
	}
	int nr_pina = 0;
	l = 0;
	cursor = 0;
	select_x = select_xc;
	select_y = select_yc;
	while(1) {
		if((PINA & (1 << PA2)) == 0) {
			nr_pina++;
		}
		if(nr_pina == 2) {
			break;
		}
	}
}

// printeaza nivelul dupa numarul de patrate si coordonatele de start
void print_lvl(int *lvl_x, int *lvl_y, int l, int *number_squares) {
	int lvl1_startx = 10 + diff_rectanglex;
	int lvl1_starty = 50;
	int lvl2_startx = 10;
	int lvl2_starty = 50;
	int rows, columns;
	int i = 0;
	if(l == 0) {
		rows = 2;
		columns = 5;
		draw_board(lvl1_startx, lvl1_starty, 5, 1);
		draw_board(lvl1_startx, lvl1_starty+diff_rectanglex, 5, 0);
	}
	else {
		rows = 2;
		columns = 7;
		draw_board(lvl2_startx, lvl1_starty, 7, 1);
		draw_board(lvl2_startx, lvl1_starty+diff_rectanglex, 7, 0);
	}
	
	if(l == 0) {
		startx = lvl1_startx;
		starty = lvl1_starty;
		*number_squares = 5;
		*lvl_x = startx;
		*lvl_y = starty;
	}
	else {
		startx = lvl2_startx;
		starty = lvl2_starty;
		*number_squares = 7;
		*lvl_x = startx;
		*lvl_y = starty;
	}
	
	revealed = calloc(2, sizeof(int *));
	for(i = 0; i < rows; i++) {
		revealed[i] = calloc(columns, sizeof(int));
	}

}

// cursorul se muta in sus in meniu
void menu_up() {
	if(cursor > 0) {
		ili9341_fillrect(select_x, select_y, rectangle_x, rectangle_y, RED);
		select_y -= 17;
		ili9341_Draw_Hollow_Circle(select_x + circle_diffx, select_y + circle_diffy, 7, BLUE);
		cursor--;
	}

}

// cursorul se muta in jos in meniu
void menu_down(int max) {
	if(cursor < max) {
		ili9341_fillrect(select_x, select_y, rectangle_x, rectangle_y, RED);
		select_y += 17;
		ili9341_Draw_Hollow_Circle(select_x + circle_diffx, select_y + circle_diffy, 7, BLUE);
		cursor++;
	}

}

// afisarea meniul depinzand de ce niveluri ai terminat
void menu(int lvl_start) {
	ili9341_clear(RED);//fill screen with a colour
	ili9341_settextsize(2.5);
	ili9341_setcursor(10, 20);
	if(lvl_start != -1){
		printf("Choose a difficulty");
	}
	if(lvl_start == 0) {
		ili9341_setcursor(70 + 50, 50 + 22);
		printf("Easy");
		ili9341_setcursor(70 + 50, 70 + 22);
		printf("Medium");
		ili9341_setcursor(70 + 50, 90 + 22);
		printf("Hard");
	} else
	if(lvl_start == 1) {
		ili9341_setcursor(70 + 50, 50 + 22);
		printf("Medium");
		ili9341_setcursor(70 + 50, 70 + 22);
		printf("Hard");
	}
	else if(lvl_start == 2)
	{
		ili9341_setcursor(70 + 50, 50 + 22);
		printf("Hard");
	}
	else {
		// win screen
		
		ili9341_setcursor(50, 10);
		printf("Congrats! You won!");
		_delay_ms(100);
		ili9341_setcursor(70, 50);
		printf("Next level?");
		ili9341_setcursor(70 + 50, 50 + 22);
		printf("Yes");
		ili9341_setcursor(70 + 50, 70 + 22);
		printf("Hall of Fame");
		l++;
		
	}
	ili9341_setcursor(10, 170);
	_delay_ms(1);
	printf("Memory Game");
	_delay_ms(2);
	ili9341_setcursor(10, 195);
	_delay_ms(1);
	printf("by Alex Maxim");
	_delay_ms(2);
	ili9341_setcursor(10, 220);
	_delay_ms(1);
	printf("Group: 332CA");
	_delay_ms(1);
}

// ecranul pentru jucatorul care a terminat un nivel
void choose_action() {
	ili9341_clear(RED);//fill screen with a colour
	ili9341_settextsize(2.5);
	ili9341_setcursor(50,80);
	printf("Congrats! You won!");
	_delay_ms(100);
	ili9341_setcursor(70, 110);
	printf("Next lvl?");
	l++;

}

// logica jocului
int main(void)
{
	DDRA = 0;
	
	// Activez PIN CHANGE INTERRUPT ENABLE pentru PCINT15:8 (portul B) si PCINT24:31 (portul D)
    PCICR |= (1 << PCIE0);
    // PD0
    PCMSK0 |= (1 << PCINT0) | (1 << PCINT1) | (1 << PCINT2) | (1 << PCINT3) | (1 << PCINT4);


	// setez pentru timer la 1 secunda ca sa am cronometru in timpul nivelului
	ICR1 = 62500;
	OCR1A = 0xF424;
	
	// activez timer-ul
	TIMSK1 = (1 << OCIE1A);
	TIMSK1 |= (1 << TOIE1);
	TCCR1B = (1 << CS12) | (1 << WGM12) | (1 << WGM13);
	
	// activez toate porturile A;
	PORTA = 255;
	stdout = & mydata;
	ili9341_init();//ili9341 setup
	ili9341_clear(RED);// umplere fundal cu rosu
	ili9341_setRotation(1);//rotire ecran
	_delay_ms(2);

	
    // afisarea initiala
	_delay_ms(2);
	ili9341_settextcolour(YELLOW,RED);
	_delay_ms(2);
		DDRD |= (1 << PD7);

		PORTD &= ~(1 << PD7);
	while(1) {
		int changed = 1;
		in_game = 0;
		// initial ma duc la meniu
		goto menu;
win_s:
		if(l == 2) {
				hall_of_fame();
				goto menu;
		}
		menu(-1);
		cursor = 0;
		while(1){
			if(changed) {
				ili9341_Draw_Hollow_Circle(cursor_menu_x + circle_diffx, cursor_menu_y + circle_diffy - 5, 7, BLUE);
				changed = 0;
			}
			
			if ((PINA & (1 << PA1)) == 0) {
				menu_up();
				ili9341_setcursor(0, 0);
				changed = 0;
			}
			_delay_ms(25);
			
			if ((PINA & (1 << PA3)) == 0) {
				menu_down(1);
				ili9341_setcursor(0, 0);
				changed = 0;
			}
			_delay_ms(25);
			if ((PINA & (1 << PA2)) == 0) {
				if(cursor == 1) {
					hall_of_fame();
					goto menu;
				}
				else {
					l++;
					goto next_lvl;
				}
				ili9341_setcursor(0, 0);
				break;
			}
		}
menu: 	cursor = 0;
		ili9341_settextsize(2.5);
		changed = 1;
		// afisez meniul
		menu(l);
		// muzica star wars
		loop();
		while(1) {
			// afisez prima data cursorul
			if(changed) {
				ili9341_Draw_Hollow_Circle(cursor_menu_x + circle_diffx, cursor_menu_y + circle_diffy - 5, 7, BLUE);
				changed = 0;
			}
			// butoanele cu care jucatorul misca cursorul in meniu, sus, jos, mijloc
			if ((PINA & (1 << PA1)) == 0) {
				menu_up();
				ili9341_setcursor(0, 0);
				changed = 0;
			}
			_delay_ms(25);
			
			if ((PINA & (1 << PA3)) == 0) {
				menu_down(2-l);
				ili9341_setcursor(0, 0);
				changed = 0;
			}
			_delay_ms(25);
			if ((PINA & (1 << PA2)) == 0) {
				l += cursor;
				ili9341_setcursor(0, 0);
				break;
			}
			_delay_ms(25);
		}
	
		ili9341_setcursor(cursorx, cursory);
		_delay_ms(1);

next_lvl: ;
		// initializare nivel nou
		ili9341_clear(RED);
		_delay_ms(100);
		seconds = 0;
		// initializez intreruperile pentru afiseaza perechilor ramase si a butoanelor
		sei();
		shape2_x = -1;
		shape2_y = -1;
		shape1_x = -1;
		shape1_y = -1;
		c2 = 0;
		c1 = 0;
		r1 = 0;
		r2 = 0;
		r = 0;
		c = 0;
		shapes_revealed = 0;
		first_time = 1;
		_delay_ms(25);
		ili9341_setcursor(30,150);
		//afisez nivelul
		print_lvl(&lvl_x, &lvl_y, l, &number_squares);
		// randomizez harta
		bind_prints(l);
		correct = 0;
		if(l == 0)
		{
			lvl_correct = 5;
		}
		else if(l == 1) {
			lvl_correct = 7;
		}
		else {
			lvl_correct = 6;
		}
		in_game = 1;
		changed = 1;
		while(1) {
			
			if(castigat) {
				//daca a terminat nivelul
				// se duce la ecranul castigatorului
				// nu se mai folosesc intreruperile
				castigat = 0;
				cli();
				goto win_s;
			}
		}
		
		startx = startxc;
		starty = startyc;
			
	}
	return 0;
}

// functie ce reda un sunet de o anumita frecventa si durata
void PLAYNOTE(float duration, float frequency)
{

	long int i, cycles;
	float half_period;
	float wavelength;

	wavelength = (1 / frequency) * 1000;
	cycles = duration / wavelength;
	half_period = wavelength / 2;

	SPEAKER_DDR |= (1 << SPEAKER_PIN);

	for (i = 0; i < cycles; i++)
	{
		_delay_ms(half_period);
		SPEAKER_PORT |= (1 << SPEAKER_PIN);
		_delay_ms(half_period);
		SPEAKER_PORT &= ~(1 << SPEAKER_PIN);
	}

	return;
}